<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xalan="http://xml.apache.org/xalan"
                xmlns:lav="http://software.in2p3.fr/lavoisier/application.xsd"
                xmlns="http://software.in2p3.fr/lavoisier/application.xsd"
                exclude-result-prefixes="xalan lav">
    <xsl:output method="xml" indent="yes" xalan:indent-amount="4"/>
    <xsl:strip-space elements="lav:*"/>

    <xsl:template match="/lav:views">
        <xsl:copy><xsl:copy-of select="@*"/>
            <xsl:choose>
                <xsl:when test="lav:view">
                    <xsl:apply-templates>
                        <xsl:with-param name="file">app/views.xml</xsl:with-param>
                        <xsl:with-param name="category">Test</xsl:with-param>
                    </xsl:apply-templates>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:for-each select="document('/views/')/properties/entry">
                        <xsl:variable name="views" select="document(@key)/lav:views"/>
                        <xsl:if test="$views/lav:processors-list">
                            <xsl:message terminate="yes">Please move your &lt;processors-list&gt; section to file engine/post-processors.xml</xsl:message>
                        </xsl:if>
                        <xsl:if test="$views/lav:authenticators-list">
                            <xsl:message terminate="yes">Please move your &lt;authenticators-list&gt; section to file service/authenticators.xml</xsl:message>
                        </xsl:if>
                        <xsl:apply-templates select="$views/lav:view">
                            <xsl:with-param name="file" select="text()"/>
                            <xsl:with-param name="category" select="substring-before(substring-after(text(),'/views/'),'.xml')"/>
                        </xsl:apply-templates>
                    </xsl:for-each>
                </xsl:otherwise>
            </xsl:choose>
            <xsl:for-each select="document('engine/views-libraries.properties')/properties/entry">
                <xsl:apply-templates select="document(@key)/lav:views/lav:view">
                    <xsl:with-param name="file" select="concat('lib/',@key,'.xml')"/>
                    <xsl:with-param name="category">Hidden</xsl:with-param>
                    <xsl:with-param name="authenticators" select="text()"/>
                </xsl:apply-templates>
            </xsl:for-each>
            <xsl:apply-templates select="document('engine/post-processors.xml')/lav:processors-list"/>
            <xsl:apply-templates select="document('service/authenticators.xml')/lav:authenticators-list"/>
        </xsl:copy>
    </xsl:template>

    <xsl:template match="lav:view">
        <xsl:param name="file"/>
        <xsl:param name="category"/>
        <xsl:param name="authenticators"/>
        <xsl:copy><xsl:copy-of select="@*"/>
            <!-- set @_file_ -->
            <xsl:if test="@_file_">
                <xsl:message terminate="yes">Attribute @_file_ is reserved for internal needs, please do not set it!</xsl:message>
            </xsl:if>
            <xsl:attribute name="_file_"><xsl:value-of select="$file"/></xsl:attribute>

            <!-- set @authenticators (if not already defined) -->
            <xsl:if test="not(@authenticators) and $authenticators">
                <xsl:attribute name="authenticators"><xsl:value-of select="$authenticators"/></xsl:attribute>
            </xsl:if>

            <!-- add <namespace> -->
            <xsl:call-template name="NAMESPACE"/>
            <xsl:apply-templates select="lav:namespace"/>

            <!-- set <category> -->
            <xsl:if test="not(lav:info)">
                <info>
                    <category>
                        <xsl:call-template name="CATEGORY">
                            <xsl:with-param name="category" select="$category"/>
                        </xsl:call-template>
                    </category>
                </info>
            </xsl:if>

            <!-- recurse -->
            <xsl:apply-templates select="*[not(self::lav:namespace)] | text() | comment()">
                <xsl:with-param name="category" select="$category"/>
            </xsl:apply-templates>
        </xsl:copy>
    </xsl:template>
    <xsl:template match="lav:pre-renderers">
        <xsl:copy><xsl:copy-of select="@*"/>
            <!-- add <namespace> -->
            <xsl:call-template name="NAMESPACE"/>

            <!-- recurse -->
            <xsl:apply-templates select="* | text() | comment()"/>
        </xsl:copy>
    </xsl:template>
    <xsl:template match="lav:info">
        <xsl:param name="category"/>
        <xsl:copy><xsl:copy-of select="@*"/>
            <!-- set <category> -->
            <xsl:if test="lav:category">
                <xsl:message terminate="yes">Category must be set in attribute /views/@name rather than in element &lt;category&gt;</xsl:message>
            </xsl:if>
            <category>
                <xsl:call-template name="CATEGORY">
                    <xsl:with-param name="category" select="$category"/>
                </xsl:call-template>
            </category>
            <xsl:apply-templates/>
        </xsl:copy>
    </xsl:template>

    <xsl:template name="CATEGORY">
        <xsl:param name="category"/>
        <xsl:choose>
            <xsl:when test="ancestor::lav:views/@name"><xsl:value-of select="ancestor::lav:views/@name"/></xsl:when>
            <xsl:otherwise><xsl:value-of select="$category"/></xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    <xsl:template name="NAMESPACE">
        <!-- convert xmlns:ns to <namespace prefix="ns"/> -->
        <!-- WARNING: running this from DOM makes order of <namespace> unpredictable -->
        <xsl:for-each select="(/* | ancestor::* | descendant-or-self::*)/namespace::*
                                    [name()!='']
                                    [.!='http://www.w3.org/XML/1998/namespace']">
            <namespace prefix="{name()}" uri="{.}"/>
        </xsl:for-each>
    </xsl:template>

    <!-- default rules -->
    <xsl:template match="comment()">
        <xsl:comment><xsl:value-of select="."/></xsl:comment>
    </xsl:template>
    <xsl:template match="*">
        <xsl:copy><xsl:copy-of select="@*"/>
            <xsl:apply-templates/>
        </xsl:copy>
    </xsl:template>
</xsl:stylesheet>